/* pathjumpui.cc
 * This file belongs to Worker, a file manager for UN*X/X11.
 * Copyright (C) 2012 Ralf Hoffmann.
 * You can contact me at: ralf@boomerangsworld.de
 *   or http://www.boomerangsworld.de/worker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "pathjumpui.hh"
#include <aguix/aguix.h>
#include <aguix/awindow.h>
#include <aguix/text.h>
#include <aguix/fieldlistview.h>
#include <aguix/button.h>
#include <aguix/solidbutton.h>
#include <aguix/util.h>
#include <aguix/choosebutton.h>
#include <aguix/cyclebutton.h>
#include "worker_locale.h"
#include "nwc_path.hh"
#include <algorithm>
#include "worker.h"
#include "deeppathstore.hh"
#include "stringmatcher_fnmatch.hh"
#include "bookmarkdbproxy.hh"
#include "persdeeppathstore.hh"
#include "prefixdb.hh"
#include "wconfig.h"
#include "stringmatcher_flexiblematch.hh"

PathJumpUI::PathJumpUI( AGUIX &aguix, DeepPathStore &path_store, Worker &worker )
  : m_aguix( aguix ),
    m_worker( worker ),
    m_path_store( path_store ),
    m_current_depth( -1 ),
    m_show_hidden_entries( false ),
    m_entry_filter_mode( SHOW_ALL ),
    m_lv_dirty( false)
{
    std::list< std::string > paths = m_path_store.getPaths();
    int max_path_components = 0;
    std::list< std::string > pers_paths = worker.getPathsPers();

    m_bookmarks = Worker::getBookmarkDBInstance().getNamesOfEntries( "" );

    paths.sort();

    DeepPathStore temp_store_paths;

    for ( std::list< std::string >::const_iterator it1 = paths.begin();
          it1 != paths.end();
          it1++ ) {
        m_entries.push_back( pathjump_entry( PATH_SHOW, -1, *it1 ) );

        temp_store_paths.storePath( *it1 );

        int slashes = std::count( it1->begin(),
                                  it1->end(),
                                  '/' );

        if ( slashes + 1 > max_path_components ) {
            max_path_components = slashes + 1;
        }
    }

    DeepPathStore temp_store;

    for ( std::list< std::string >::const_iterator it1 = m_bookmarks.begin();
          it1 != m_bookmarks.end();
          it1++ ) {
        //TODO it would be slightly faster to check if path is already
        //     stored in temp_store_paths but there is no
        //     corresponding function at the moment
        bool changed = false;

        temp_store_paths.storePath( *it1, &changed );

        if ( changed ) {
            temp_store.storePath( *it1 );
        }
    }

    for ( std::list< std::string >::const_iterator it1 = pers_paths.begin();
          it1 != pers_paths.end();
          it1++ ) {
        //TODO it would be slightly faster to check if path is already
        //     stored in temp_store_paths but there is no
        //     corresponding function at the moment
        bool changed = false;

        temp_store_paths.storePath( *it1, &changed );

        if ( changed ) {
            temp_store.storePath( *it1 );
        }
    }

    std::list< std::string > temp_paths = temp_store.getPaths();

    for ( std::list< std::string >::const_iterator it1 = temp_paths.begin();
          it1 != temp_paths.end();
          it1++ ) {
        m_entries.push_back( pathjump_entry( BOOKMARK_HIDE, -1, *it1 ) );

        int slashes = std::count( it1->begin(),
                                  it1->end(),
                                  '/' );

        if ( slashes + 1 > max_path_components ) {
            max_path_components = slashes + 1;
        }
    }

    // without leading slash there could be this many components,
    // add one for the initial slash
    max_path_components++;

    m_win = std::auto_ptr<AWindow>( new AWindow( &m_aguix,
                                                 0, 0,
                                                 400, 400,
                                                 0, catalog.getLocaleCom( 60 ) ) );
    m_win->create();

    m_co1 = m_win->setContainer( new AContainer( m_win.get(), 1, 6 ), true );
    m_co1->setMaxSpace( 5 );
    
    AContainer *co1_2 = m_co1->add( new AContainer( m_win.get(), 2, 1 ), 0, 0 );
    co1_2->setMaxSpace( 5 );
    co1_2->setBorderWidth( 0 );
    co1_2->add( new Text( &m_aguix, 0, 0, catalog.getLocale( 758 ), 1 ),
                0, 0, AContainer::CO_FIX );
    m_infixtext = static_cast<Text*>( co1_2->add( new Text( &m_aguix, 0, 0, "", 1 ),
                                                  1, 0, AContainer::CO_INCW ) );


    AContainer *co1_4 = m_co1->add( new AContainer( m_win.get(), 2, 1 ), 0, 1 );
    co1_4->setMaxSpace( 5 );
    co1_4->setBorderWidth( 0 );
    co1_4->add( new Text( &m_aguix, 0, 0, catalog.getLocale( 980 ), 1 ),
                0, 0, AContainer::CO_FIX );
    m_filter_mode_cyb = (CycleButton*)co1_4->add( new CycleButton( &m_aguix, 0, 0, 50, 1, 0, 0 ),
                                                  1, 0, AContainer::CO_INCWNR );

    m_filter_mode_cyb->addOption( catalog.getLocale( 981 ) );
    m_filter_mode_cyb->addOption( catalog.getLocale( 982 ) );
    m_filter_mode_cyb->addOption( catalog.getLocale( 983 ) );

    m_filter_mode_cyb->resize( m_filter_mode_cyb->getMaxSize(),
                               m_filter_mode_cyb->getHeight() );
    co1_4->readLimits();

    m_show_all_cb = (ChooseButton*)m_co1->add( new ChooseButton( &m_aguix, 0, 0,
                                                                 false,
                                                                 catalog.getLocale( 977 ),
                                                                 LABEL_RIGHT, 1, 0 ),
                                               0, 2, AContainer::CO_INCWNR );

    m_breadcrumb_co = m_co1->add( new AContainer( m_win.get(), max_path_components + 1, 1 ), 0, 3 );
    m_breadcrumb_co->setMaxSpace( 5 );

    for ( int i = 0; i < max_path_components; i++ ) {
        Button *b = (Button*)m_breadcrumb_co->add( new Button( &m_aguix,
                                                               0,
                                                               0,
                                                               "",
                                                               1,
                                                               0,
                                                               0 ), i, 0, AContainer::CO_INCW );
        m_breadcrumb_buttons.push_back( b );
        b->setAcceptFocus( false );
        b->hide();
    }

    m_lv = static_cast<FieldListView*>( m_co1->add( new FieldListView( &m_aguix, 
                                                                     0, 0, 
                                                                     400, 300, 0 ),
                                                  0, 4, AContainer::CO_MIN ) );
    m_lv->setNrOfFields( 2 );
    m_lv->setShowHeader( true );
    m_lv->setFieldText( 1, catalog.getLocale( 809 ) );
    m_lv->setHBarState( 2 );
    m_lv->setVBarState( 2 );
    m_lv->setAcceptFocus( true );
    m_lv->setDisplayFocus( true );
    m_lv->setFieldText( 0, catalog.getLocale( 978 ) );
    m_lv->setGlobalFieldSpace( 5 );

    AContainer *co1_3 = m_co1->add( new AContainer( m_win.get(), 3, 1 ), 0, 5 );
    co1_3->setMinSpace( 5 );
    co1_3->setMaxSpace( -1 );
    co1_3->setBorderWidth( 0 );
    m_okb = (Button*)co1_3->add( new Button( &m_aguix,
                                             0,
                                             0,
                                             catalog.getLocale( 761 ),
                                             1,
                                             0,
                                             0 ), 0, 0, AContainer::CO_FIX );
    m_cleanupb = (Button*)co1_3->add( new Button( &m_aguix,
                                                  0,
                                                  0,
                                                  catalog.getLocale( 984 ),
                                                  1,
                                                  0,
                                             0 ), 1, 0, AContainer::CO_FIX );
    m_cancelb = (Button*)co1_3->add( new Button( &m_aguix,
                                                 0,
                                                 0,
                                                 catalog.getLocale( 633 ),
                                                 1,
                                                 0,
                                                 0 ), 2, 0, AContainer::CO_FIX );

    m_win->contMaximize( true );
    m_win->setDoTabCycling( true );

    std::string cfgfile = Worker::getWorkerConfigDir();
#ifdef USEOWNCONFIGFILES
    cfgfile = NWC::Path::join( cfgfile, "directory-history.prob2" );
#else
    cfgfile = NWC::Path::join( cfgfile, "directory-history.prob" );
#endif

    // make this instance static in this class or worker to avoid
    // reparsing it
    m_pdb = std::auto_ptr< PrefixDB >( new PrefixDB( cfgfile ) );
}

PathJumpUI::~PathJumpUI()
{
}

int PathJumpUI::mainLoop()
{
    m_lv_dirty = true;

    showData();

    highlight_best_hit( m_dirname );

    maximizeWin();
    m_win->show();

    m_lv->takeFocus();

    AGMessage *msg;
    int endmode = 0;

    for ( ; endmode == 0; ) {
        msg = m_aguix.WaitMessage( NULL );
        if ( msg != NULL ) {
            switch ( msg->type ) {
              case AG_CLOSEWINDOW:
                  endmode = -1;
                  break;
              case AG_BUTTONCLICKED:
                  if ( msg->button.button == m_okb ) {
                      endmode = 1;
                  } else if ( msg->button.button == m_cleanupb ) {
                      cleanup();
                      endmode = -1;
                  } else if ( msg->button.button == m_cancelb ) {
                      endmode = -1;
                  } else {
                      for ( unsigned int i = 0; i < m_breadcrumb_buttons.size(); i++ ) {
                          if ( m_breadcrumb_buttons[i] == msg->button.button ) {
                              m_current_depth = i;
                              endmode = 1;
                          }
                      }
                  }
                  break;
              case AG_KEYPRESSED:
                  if ( msg->key.key == XK_BackSpace ) {
                      if ( KEYSTATEMASK( msg->key.keystate ) == ShiftMask ) {
                          m_filter = "";
                          apply_filter( m_filter );
                          showData();
                      } else {
                          std::string cur_infix = m_filter;
                          
                          if ( cur_infix.length() > 0 ) {
                              cur_infix.resize( cur_infix.length() - 1 );
                              m_filter = cur_infix;
                              apply_filter( m_filter );
                              showData();
                          }
                      }
                  } else if ( msg->key.key == XK_Return ) {
                      endmode = 1;
                  } else if ( msg->key.key == XK_Escape ) {
                      endmode = -1;
                  } else if ( msg->key.key == XK_Left ) {
                      if ( ! m_current_components.empty() ) {
                          if ( m_current_depth < 0 ) {
                              m_current_depth = m_current_components.size();
                          } else if ( m_current_depth > 0 ) {
                              if ( m_current_depth > (int)m_current_components.size() &&
                                   m_current_components.size() >= 1 ) {
                                  m_current_depth = m_current_components.size();
                              }

                              m_current_depth--;
                          }

                          if ( m_entry_filter_mode != SHOW_ALL ) {
                              post_filtering();
                              showData();
                          } else {
                              update_breadcrumb();
                          }
                      }
                  } else if ( msg->key.key == XK_Right ) {
                      if ( ! m_current_components.empty() ) {
                          if ( m_current_depth < 0 ) {
                              m_current_depth = 0;
                          } else if ( m_current_depth <= (int)( m_current_components.size() - 1 ) ) {
                              m_current_depth++;
                          }

                          if ( m_entry_filter_mode != SHOW_ALL ) {
                              post_filtering();
                              showData();
                          } else {
                              update_breadcrumb();
                          }
                      }
                  } else if ( msg->key.key == XK_Up ) {
                  } else if ( msg->key.key == XK_Down ) {
                  } else if ( msg->key.key == XK_Next ) {
                  } else if ( msg->key.key == XK_Prior ) {
                  } else if ( msg->key.key == XK_Home ) {
                      // key is grabbed by listview so it has no effect at the moment
                      if ( ! m_current_components.empty() ) {
                          m_current_depth = 0;
                          update_breadcrumb();
                      }
                  } else if ( msg->key.key == XK_End ) {
                      // key is grabbed by listview so it has no effect at the moment
                      if ( ! m_current_components.empty() ) {
                          m_current_depth = m_current_components.size();
                          update_breadcrumb();
                      }
                  } else if ( IsModifierKey( msg->key.key ) ||
                              IsCursorKey( msg->key.key ) ||
                              IsPFKey( msg->key.key ) ||
                              IsFunctionKey( msg->key.key ) ||
                              IsMiscFunctionKey( msg->key.key ) ||
                              ( ( msg->key.key >= XK_BackSpace ) && ( msg->key.key <= XK_Escape ) ) ) {
                  } else if ( strlen( msg->key.keybuf ) > 0 ) {
                      if ( msg->key.key == XK_a &&
                           KEYSTATEMASK( msg->key.keystate ) == ControlMask ) {
                          m_show_hidden_entries = ! m_show_hidden_entries;
                          m_show_all_cb->setState( m_show_hidden_entries );
                          apply_filter( m_filter );
                          showData();
                      } else if ( msg->key.key == XK_f &&
                           KEYSTATEMASK( msg->key.keystate ) == ControlMask ) {
                          switch ( m_entry_filter_mode ) {
                              case SHOW_ONLY_SUBDIRS:
                                  m_entry_filter_mode = SHOW_ONLY_DIRECT_SUBDIRS;
                                  m_filter_mode_cyb->setOption( 2 );
                                  break;
                              case SHOW_ONLY_DIRECT_SUBDIRS:
                                  m_entry_filter_mode = SHOW_ALL;
                                  m_filter_mode_cyb->setOption( 0 );
                                  break;
                              default:
                                  m_entry_filter_mode = SHOW_ONLY_SUBDIRS;
                                  m_filter_mode_cyb->setOption( 1 );
                                  break;
                          }
                          apply_filter( m_filter );
                          showData();
                      } else {
                          std::string cur_infix = m_filter;
                          cur_infix += msg->key.keybuf;
                          if ( apply_filter( cur_infix ) > 0 ) {
                              m_filter = cur_infix;
                              apply_filter( m_filter );
                              showData();
                          }
                      }
                  }
                  break;
              case AG_FIELDLV_DOUBLECLICK:
                  if ( msg->fieldlv.lv == m_lv ) {
                      // double click in lv, actual element is unimportant here
                      endmode = 1;
                  }
                  break;
                case AG_FIELDLV_ONESELECT:
                case AG_FIELDLV_MULTISELECT:
                    if ( msg->fieldlv.lv == m_lv ) {
                        try {
                            m_current_entry = m_entry_fullnames.at( m_lv->getActiveRow() );
                            set_entry( m_current_entry );
                            showData();
                        } catch (...) {
                        }
                    }
                    break;
                case AG_CHOOSECLICKED:
                    if ( msg->choose.button == m_show_all_cb ) {
                        m_show_hidden_entries = msg->choose.state;
                        apply_filter( m_filter );
                        showData();
                    }
                    break;
                case AG_CYCLEBUTTONCLICKED:
                    if ( msg->cyclebutton.cyclebutton == m_filter_mode_cyb ) {
                        switch ( msg->cyclebutton.option ) {
                            case 1:
                                m_entry_filter_mode = SHOW_ONLY_SUBDIRS;
                                break;
                            case 2:
                                m_entry_filter_mode = SHOW_ONLY_DIRECT_SUBDIRS;
                                break;
                            default:
                                m_entry_filter_mode = SHOW_ALL;
                                break;
                        }
                        apply_filter( m_filter );
                        showData();
                    }
                    break;
            }
            m_aguix.ReplyMessage( msg );
        }
    }

    m_selected_path = "";

    if ( endmode == 1 ) {
        std::string res = "";

        if ( m_current_depth <= 0 ) {
            res = "/";
        } else {
            for ( unsigned int i = 0; i < m_current_components.size(); i++ ) {
                if ( (int)i >= m_current_depth ) break;

                res += "/";
                res += m_current_components[i];
            }
        }

        m_selected_path = res;

        // check if this path is allowed to be tracked, than put its
        // access into the prefixdb
        if ( ! m_filter.empty() && ! wconfig->getPathJumpAllowDirs().empty() ) {
            const std::list< std::string > &l = wconfig->getPathJumpAllowDirs();

            for ( std::list< std::string >::const_iterator it1 = l.begin();
                  it1 != l.end();
                  it1++ ) {
                if ( AGUIXUtils::starts_with( m_selected_path, *it1 ) ) {
                    m_pdb->pushAccess( m_filter, m_selected_path );
                    break;
                }
            }
        }
    }

    m_win->hide();
    
    return endmode;
}

void PathJumpUI::showData()
{
    m_infixtext->setText( m_filter.c_str() );

    if ( m_lv_dirty ||
         m_entry_filter_mode == SHOW_ONLY_DIRECT_SUBDIRS ) {

        m_lv->setSize( 0 );
        m_entry_fullnames.resize( 0 );
    
        int size = 0, pos = 0;

        std::string prefix_filter;

        if ( m_entry_filter_mode == SHOW_ONLY_SUBDIRS ||
             m_entry_filter_mode == SHOW_ONLY_DIRECT_SUBDIRS ) {
            int i = 0;

            if ( m_current_depth > 0 ) {
                for ( std::vector< std::string >::const_iterator it1 = m_current_components.begin();
                      it1 != m_current_components.end();
                      it1++, i++ ) {
                    prefix_filter += "/";
                    prefix_filter += *it1;
                    if ( i + 1 >= m_current_depth ) break;
                }
            }
        }

        size_t l = prefix_filter.length();
        bool current_entry_found = false;

        for ( std::list< pathjump_entry >::const_iterator it1 = m_entries.begin();
              it1 != m_entries.end();
              it1++, pos++ ) {
            if ( ( it1->m_type == PATH_SHOW ||
                   it1->m_type == BOOKMARK_SHOW ) &&
                 it1->m_post_filtering ) {
                int row = m_lv->addRow();

                m_entry_fullnames.push_back( it1->m_path );

                if ( m_entry_filter_mode == SHOW_ONLY_DIRECT_SUBDIRS ) {
                    size_t p2 = it1->m_path.find( '/', l + 1 );
                    if ( p2 != std::string::npos ) {
                        std::string s( it1->m_path, 0, p2 );

                        m_lv->setText( row, 1, s );
                    } else {
                        m_lv->setText( row, 1, it1->m_path );
                    }
                } else {
                    m_lv->setText( row, 1, it1->m_path );
                }

                if ( it1->m_blockcount >= 0 && ! m_filter.empty() ) {
                    std::string qual = AGUIXUtils::formatStringToString( "%3.0f%%",
                                                                         100.0 / (double)( it1->m_blockcount + 1 ) );
                    m_lv->setText( row, 0, qual );
                }

                m_lv->setPreColors( row, FieldListView::PRECOLOR_ONLYACTIVE );

                if ( m_current_entry == it1->m_path ) {
                    m_lv->setActiveRow( row );
                    current_entry_found = true;
                }

                size++;
            }
        }

        if ( ! current_entry_found ) {
            for ( int row = 0; row < m_lv->getElements(); row++ ) {
                if ( AGUIXUtils::starts_with( m_current_entry,
                                              m_lv->getText( row, 1 ) ) ) {
                    m_lv->setActiveRow( row );
                    break;
                }
            }
        }

        m_lv->showActive();
        m_lv->redraw();
    }

    update_breadcrumb();

    m_lv_dirty = false;
}

std::string PathJumpUI::getSelectedPath()
{
    return m_selected_path;
}

void PathJumpUI::setCurrentDirname( const std::string &dirname )
{
    m_dirname = dirname;
}

void PathJumpUI::setCurrentBasename( const std::string &basename )
{
    m_basename = basename;
}

void PathJumpUI::maximizeWin()
{
    m_lv->maximizeX();
    m_lv->maximizeY();
    int my_w = m_lv->getWidth() + 10;
    int my_h = m_lv->getHeight() + 10;

    if ( my_w < 400 ) my_w = 400;
    if ( my_h < 300 ) my_h = 300;

    int rx, ry, rw, rh;

    m_aguix.getLargestDimensionOfCurrentScreen( &rx, &ry,
                                                &rw, &rh );

    int mw = rw * 80 / 100;
    int mh = rh * 80 / 100;
    m_co1->resize( mw, mh );
    m_co1->rearrange();

    if ( my_w < m_lv->getWidth() ) {
        m_co1->setMinWidth( my_w, 0, 4 );
    } else {
        m_co1->setMinWidth( m_lv->getWidth(), 0, 4 );
    }
    if ( my_h < m_lv->getHeight() ) {
        m_co1->setMinHeight( my_h, 0, 4 );
    } else {
        m_co1->setMinHeight( m_lv->getHeight(), 0, 4 );
    }
    m_win->contMaximize( true );
}

void PathJumpUI::highlight_best_hit( const std::string &filter )
{
    find_entry( filter, "" );

    showData();
}

int PathJumpUI::apply_filter( const std::string &filter )
{
    int matches = 0;

    StringMatcherFlexibleMatch flmatch;
    StringMatcherFNMatch match;
    bool use_flexible = true;

    flmatch.setMatchString( filter );
    flmatch.setMatchCaseSensitive( false );

    if ( filter.find( "*" ) != std::string::npos ) {
        std::string real_filter = "*";
        real_filter += filter;
        real_filter += "*";

        match.setMatchString( real_filter );
        match.setMatchCaseSensitive( false );

        use_flexible = false;
    }

    int exact_path_matches = 0, exact_bookmark_matches = 0,
        path_matches = 0, bookmark_matches = 0;
    bool prefer_exact_matches = false;

    for ( std::list< pathjump_entry >::iterator it1 = m_entries.begin();
          it1 != m_entries.end();
          it1++ ) {

        if ( use_flexible ) {
            it1->m_blockcount = flmatch.countNonMatchingBlocks( it1->m_path );
        } else {
            it1->m_blockcount = match.match( it1->m_path ) ? 0 : -1;
        }

        if ( it1->m_blockcount == 0 ) {
            if ( it1->m_type == PATH_SHOW ||
                 it1->m_type == PATH_HIDE ) {
                exact_path_matches++;
                path_matches++;
            } else {
                exact_bookmark_matches++;
                bookmark_matches++;
            }
        } else if ( it1->m_blockcount > 0 ) {
            if ( it1->m_type == PATH_SHOW ||
                 it1->m_type == PATH_HIDE ) {
                path_matches++;
            } else {
                bookmark_matches++;
            }
        }
    }

    m_entries.sort();

    for ( std::list< pathjump_entry >::iterator it1 = m_entries.begin();
          it1 != m_entries.end();
          it1++ ) {
        if ( it1->m_type == PATH_SHOW ||
             it1->m_type == PATH_HIDE ) {
            if ( it1->m_blockcount == 0 ||
                 ( it1->m_blockcount > 0 &&
                   ( ! prefer_exact_matches || ( prefer_exact_matches && exact_path_matches == 0 ) ) ) ) {
                it1->m_type = PATH_SHOW;
                matches++;
            } else {
                it1->m_type = PATH_HIDE;
            }
        } else {
            if ( ( path_matches == 0 && ! filter.empty() ) ||
                 m_show_hidden_entries == true ) {
                if ( it1->m_blockcount == 0 ||
                     ( it1->m_blockcount > 0 &&
                       ( ! prefer_exact_matches || ( prefer_exact_matches && exact_bookmark_matches == 0 ) ) ) ) {
                    it1->m_type = BOOKMARK_SHOW;
                    matches++;
                } else {
                    it1->m_type = BOOKMARK_HIDE;
                }
            } else {
                    it1->m_type = BOOKMARK_HIDE;
            }
        }
    }

    find_entry( m_current_entry, filter );

    m_lv_dirty = true;

    return matches;
}

int PathJumpUI::find_entry( const std::string &name, const std::string &filter )
{
    std::string entry_to_find;
    bool first_found = false;
    std::string best_hit;

    if ( ! filter.empty() ) {
        best_hit = entry_to_find = m_pdb->getBestHit( filter );
    }

    if ( entry_to_find.empty() ) {
        entry_to_find = name;
    }

    for ( std::list< pathjump_entry >::iterator it1 = m_entries.begin();
          it1 != m_entries.end();
          it1++ ) {
        if ( it1->m_type == PATH_SHOW ||
             it1->m_type == BOOKMARK_SHOW ) {
            if ( AGUIXUtils::starts_with( it1->m_path,
                                          entry_to_find ) ||
                 first_found == false ) {
                m_current_entry = it1->m_path;

                if ( first_found ) {
                    break;
                } else {
                    first_found = true;
                }
            }
        }
    }


    if ( ! best_hit.empty() ) {
        std::string f = best_hit;
        f += "*";
        
        m_current_depth = find_best_matching_depth( m_current_entry, f );
    } else {
        if ( ! filter.empty() ) {
            m_current_depth = find_best_matching_depth( m_current_entry, filter );
        } else {
            // workaround to adjust the depth to the name argument
            m_current_depth = find_best_matching_depth( m_current_entry, name + "*" );
        }
    }

    if ( m_entry_filter_mode != SHOW_ALL ) {
        m_lv_dirty = true;
    }

    post_filtering();

    return 0;
}

int PathJumpUI::set_entry( const std::string &name )
{
    bool found = false;

    for ( std::list< pathjump_entry >::iterator it1 = m_entries.begin();
          it1 != m_entries.end();
          it1++ ) {
        if ( ( it1->m_type == PATH_SHOW ||
               it1->m_type == BOOKMARK_SHOW ) &&
             it1->m_path == name ) {
            m_current_entry = it1->m_path;
            found = true;
            break;
        }
    }

    if ( found ) {
        if ( ! m_filter.empty() &&
             m_entry_filter_mode == SHOW_ALL ) {
            m_current_depth = find_best_matching_depth( m_current_entry, m_filter );
        }
    } else {
        find_entry( name, m_filter );
    }

    if ( m_entry_filter_mode != SHOW_ALL ) {
        m_lv_dirty = true;
    }

    post_filtering();

    return 0;
}

int PathJumpUI::post_filtering()
{
    std::string prefix_filter;

    build_current_path_components( m_current_entry );

    if ( m_entry_filter_mode != SHOW_ONLY_SUBDIRS &&
         m_entry_filter_mode != SHOW_ONLY_DIRECT_SUBDIRS ) {

        for ( std::list< pathjump_entry >::iterator it1 = m_entries.begin();
              it1 != m_entries.end();
              it1++ ) {
            if ( it1->m_type == PATH_SHOW ||
                 it1->m_type == BOOKMARK_SHOW ) {

                if ( ! it1->m_post_filtering ) m_lv_dirty = true;

                it1->m_post_filtering = true;
            } else {

                if ( it1->m_post_filtering ) m_lv_dirty = true;

                it1->m_post_filtering = false;
            }
        }

        return 0;
    }

    if ( m_current_depth > 0 ) {
        int i = 0;
        for ( std::vector< std::string >::const_iterator it1 = m_current_components.begin();
              it1 != m_current_components.end();
              it1++, i++ ) {
            prefix_filter += "/";
            prefix_filter += *it1;
            if ( i + 1 >= m_current_depth ) break;
        }
    }

    size_t l = prefix_filter.length();
    std::map< std::string, bool > entry_visible;

    for ( std::list< pathjump_entry >::iterator it1 = m_entries.begin();
          it1 != m_entries.end();
          it1++ ) {

        bool new_state = false;

        if ( it1->m_type == PATH_SHOW ||
             it1->m_type == BOOKMARK_SHOW ) {
            new_state = true;
        }

        if ( ! prefix_filter.empty() &&
             ! AGUIXUtils::starts_with( it1->m_path,
                                        prefix_filter ) ) {
            new_state = false;
        }

        if ( m_entry_filter_mode == SHOW_ONLY_DIRECT_SUBDIRS &&
             new_state ) {

            size_t p2 = it1->m_path.find( '/', l + 1 );
            std::string s;
            if ( p2 != std::string::npos ) {
                s = std::string( it1->m_path, 0, p2 );
            } else {
                s = it1->m_path;
            }

            if ( entry_visible[s] == true ) {
                new_state = false;
            } else {
                entry_visible[s] = true;
            }
        }

        if ( it1->m_post_filtering != new_state ) m_lv_dirty = true;

        it1->m_post_filtering = new_state;
    }

    return 0;
}

int PathJumpUI::update_breadcrumb()
{
    unsigned int i = 0;
    unsigned int last_component = m_current_components.size();

    m_breadcrumb_buttons[i]->setText( 0, "/" );

    if ( m_current_depth == (int)i ||
         ( i == last_component && m_current_depth > (int)last_component ) ) {
        m_breadcrumb_buttons[i]->setFG( 0, 2 );
        m_breadcrumb_buttons[i]->setBG( 0, 3 );

        m_breadcrumb_co->setMinWidth( m_breadcrumb_buttons[i]->getMaximumWidth(),
                                      i, 0 );
    } else {
        m_breadcrumb_buttons[i]->setFG( 0, 1 );
        m_breadcrumb_buttons[i]->setBG( 0, 0 );

        m_breadcrumb_co->setMinWidth( 10,
                                      i, 0 );
    }

    m_breadcrumb_buttons[i]->show();

    m_breadcrumb_co->setMaxWidth( m_breadcrumb_buttons[i]->getMaximumWidth(),
                                  i, 0 );
    i++;

    for ( std::vector< std::string >::const_iterator it1 = m_current_components.begin();
          it1 != m_current_components.end();
          it1++ ) {
        if ( ! it1->empty() ) {
            m_breadcrumb_buttons[i]->setText( 0, it1->c_str() );

            if ( m_current_depth == (int)i ||
                 ( i == last_component && m_current_depth > (int)last_component ) ) {
                m_breadcrumb_buttons[i]->setFG( 0, 2 );
                m_breadcrumb_buttons[i]->setBG( 0, 3 );

                m_breadcrumb_co->setMinWidth( m_breadcrumb_buttons[i]->getMaximumWidth(),
                                              i, 0 );
            } else {
                m_breadcrumb_buttons[i]->setFG( 0, 1 );
                m_breadcrumb_buttons[i]->setBG( 0, 0 );

                m_breadcrumb_co->setMinWidth( 10,
                                              i, 0 );
            }

            m_breadcrumb_buttons[i]->show();

            m_breadcrumb_co->setMaxWidth( m_breadcrumb_buttons[i]->getMaximumWidth(),
                                          i, 0 );
            i++;
        }
    }

    for ( ; i < m_breadcrumb_buttons.size(); i++ ) {
        m_breadcrumb_buttons[i]->hide();
        m_breadcrumb_co->setMaxWidth( 0,
                                      i, 0 );
        m_breadcrumb_co->setMinWidth( 0,
                                      i, 0 );
    }

    m_win->updateCont();

    return 0;
}

int PathJumpUI::find_best_matching_depth( const std::string &filter )
{
    if ( m_lv->isValidRow( m_lv->getActiveRow() ) == true ) {
        return find_best_matching_depth( m_lv->getText( m_lv->getActiveRow(), 1 ),
                                         filter );
    }
    return -1;
}

int PathJumpUI::find_best_matching_depth( const std::string &path,
                                          const std::string &filter )
{
    std::vector< std::string > components;
    StringMatcherFNMatch matcher;
    int depth = -1;

    if ( filter.empty() ) return -1;

    AGUIXUtils::split_string( components, path, '/' );

    matcher.setMatchString( filter );
    matcher.setMatchCaseSensitive( false );

    if ( ! matcher.match( path ) ) {
        std::string flexible_filter;

        if ( filter[0] != '*' ) {
            flexible_filter = "*";
        }

        flexible_filter += StringMatcherFNMatch::convertMatchStringToFlexibleMatch( filter ) + "*";

        matcher.setMatchString( flexible_filter );
        matcher.setMatchCaseSensitive( false );
    }

    int i = 0;
    std::string prefix_path = "";

    for ( std::vector< std::string >::const_iterator it1 = components.begin();
          it1 != components.end();
          it1++ ) {
        if ( ! it1->empty() ) {
            prefix_path += "/" + *it1;

            if ( matcher.match( prefix_path ) ) {
                depth = i + 1;
                break;
            }

            i++;
        }
    }

    return depth;
}

int PathJumpUI::build_current_path_components()
{
    if ( m_lv->isValidRow( m_lv->getActiveRow() ) == true ) {
        return build_current_path_components( m_lv->getText( m_lv->getActiveRow(), 1 ) );
    }
    return 1;
}

int PathJumpUI::build_current_path_components( const std::string &path )
{
    std::vector< std::string > components;

    m_current_components.clear();

    AGUIXUtils::split_string( components, path, '/' );

    if ( components.size() >= m_breadcrumb_buttons.size() ) {
        // something wrong here, shouldn't happen
        return 1;
    }

    for ( std::vector< std::string >::const_iterator it1 = components.begin();
          it1 != components.end();
          it1++ ) {
        if ( ! it1->empty() ) {
            m_current_components.push_back( *it1 );
        }
    }

    return 0;
}

int PathJumpUI::cleanup()
{
    std::string msg, bt;

    std::list< std::string > temp_paths = m_worker.getPathsPers(),
        non_existing_entries;

    int non_existing = 0, sum = 0;

    m_win->setCursor( AGUIX::WAIT_CURSOR );
    m_aguix.Flush();

    for ( std::list< std::string >::const_iterator it1 = temp_paths.begin();
          it1 != temp_paths.end();
          it1++ ) {
        if ( ! NWC::FSEntry( *it1 ).entryExists() ) {
            non_existing_entries.push_back( *it1 );
            non_existing++;
        }

        sum++;
    }

    m_win->unsetCursor();

    if ( non_existing > 0 ) {
        msg = AGUIXUtils::formatStringToString( catalog.getLocale( 985 ),
                                                non_existing, sum );

        bt = catalog.getLocale( 11 );
        bt += "|";
        bt += catalog.getLocale( 8 );

        int res = m_win->request( catalog.getLocale( 123 ),
                                  msg.c_str(),
                                  bt.c_str() );

        if ( res == 0 ) {
            for ( std::list< std::string >::const_iterator it1 = non_existing_entries.begin();
                  it1 != non_existing_entries.end();
                  it1++ ) {
                m_worker.removePathPers( *it1 );
            }
        }
    } else {
        m_win->request( catalog.getLocale( 123 ),
                        catalog.getLocale( 986 ),
                        catalog.getLocale( 633 ) );
    }

    return 0;
}
